(* module Types

   The purpose of this module is to provide several different
   implementations of the ENV signature declared in
   interfaces.mli. You can choose between Env1 and Env2
   without affecting the program's behavior.

   Env0 is a dummy implementation that has the right interface,
   but doesn't save any bindings. This won't affect the
   VA/substitute-and-repeat interpreter, except that toplevel bindings
   (`let x = ... end`) and `symbol` declarations will no longer
   have any effect. (The parser will still accept them; they'll
   just be inert. Env0 is not suitable for use with the
   VB/environment-based interpreter, which needs a real
   implementation of environments.
 
   Env1 implements the environment as a list of pairs, where
   the first element in the pair is the assigned identifier, and
   the second element is the result that it's bound to, wrapped up
   in the `binding` datatype from interface.mli:
 
     (* Wrappers around what we save in environments. *)
     type binding =
       (* not used in VA/substitute-and-repeat interpreter *)
       | LocalBind of bound
       (* when `let x = ... end` was declared at toplevel *)
       | TopBind of bound
       (* when `symbol x` was declared at toplevel *)
       | SymbolBind
 
   Env2 instead implements the environment as a function from
   identifiers to results (wrapped up in the `binding` datatype,
   as before.)
 
   At the end of this file is the line:
 
     include Env1
 
   Just change this to whichever of these implementations you'd
   like to use.
 
   Because of some nasty details that we won't explain, most of
   the types declared in interfaces.mli are exported from this
   module, rather than from there. This is why other files in the
   program say `open Types` at their top, rather than `open
   Interface`. A few files say both.
*)

open Interface

module Env0 : ENV = struct
  type env = unit
  (* next line is trickery, see below *)
     module rec S : SYNTAX with type tmp_env = env = S include S
  let empty = ()
  let push (ident : identifier) binding () = ()
  let rec lookup (sought_ident : identifier) () = None
  let version = "null environments"
end (* module Env0 *)

module Env1 : ENV = struct
  (* We want to declare:
         type env = (identifier * binding) list
     where those are the same identifier and binding as in
     Interface.SYNTAX however we have to resort to some trickery to
     do it, which is confined to the next line *)
     module type S = sig type tmp_binding type env = (identifier * tmp_binding) list include SYNTAX with type tmp_env = env end module rec S : S with type tmp_binding = S.binding = S include S
  (* end of trickery; see http://stackoverflow.com/questions/29115356 for discussion *)

  let empty = []
  let push (ident : identifier) binding env = (ident,binding)::env
  let rec lookup (sought_ident : identifier) = function
    | [] -> None
    | (ident,binding)::_ when ident = sought_ident -> Some binding
    | _::env' -> lookup sought_ident env'
  let version = "list environments"
end (* module Env1 *)

module Env2 : ENV = struct
  (* We want to declare:
         type env = identifier -> binding option
     but as before, we need to resort to some trickery, which
     is confined to the next line *)
     module type S = sig type tmp_binding type env = identifier -> tmp_binding option include SYNTAX with type tmp_env = env end module rec S : S with type tmp_binding = S.binding = S include S
  (* end of trickery *)

  let empty = fun _ -> None
  let push (ident : identifier) binding env =
    fun (sought_ident : identifier) ->
      if ident = sought_ident
      then Some binding
      else env sought_ident
  let lookup sought_ident env = env sought_ident
  let version = "functional environments"
end (* module Env2 *)

(* Any of the above choices should be valid. *)
include Env1
